import os
from pathlib import Path
from typing import AnyStr

current_dir = Path(__file__).resolve().parent

if current_dir.name == 'src':
    current_dir = current_dir.parent

examples_dir = current_dir / "src/examples"
direction_dir = current_dir / "docs/examples"


def git_add():
    """Add new files to git"""
    os.system(f"git add {direction_dir}")


def read(name: Path) -> AnyStr:
    """Reads the content of doc file"""
    with name.open("r") as f:
        return f.read()


def make_markdown(doc: AnyStr) -> AnyStr:
    """Make markdown doc format"""
    return f"```python\n{doc}\n```"


def record(name: str, content: str):
    """Record the content of doc file"""
    file = Path(direction_dir / f"{name.removesuffix('.py')}.md")
    dir_path = file.parent

    dir_path.mkdir(exist_ok=True)
    with file.open("w") as f:
        f.write(content)

    return 1


def skip_file(file_name: AnyStr) -> bool:
    """Check if a file should be skipped"""
    return file_name == '__init__.py' or '__pycache__' in file_name or ".md" in file_name


def cli(dir_path: Path, recorded=0):
    """CLI command readdocs"""
    file: Path
    for file in dir_path.iterdir():
        if skip_file(file.name):
            continue

        if file.is_dir():
            return cli(file, recorded)

        content = read(file)
        dir_name = f"{dir_path.name}/" if dir_path is not examples_dir else ''
        recorded += record(f"{dir_name}{file.name}", make_markdown(content))

    print(f'Recorded: {recorded} files')

def main():
    cli(examples_dir)
    git_add()


if __name__ == '__main__':
    main()



import asyncio
import dataclasses
import logging

from examples.books.messages import CreateBook, BookCreated, BookQuery
from examples.books.models import Book
from iambus import dispatcher as dp
from iambus.core import signals
from iambus.core.dependency.providers import Singleton


async def get_storage():
    """Dependency injection of storage."""
    return {}


# somewhere in your code
@dp.commands.register(CreateBook, response_event=BookCreated)
# ensure Singleton dependency name is the same that in handler's args
async def create_book_handler(command: CreateBook) -> BookCreated:
    """Create book"""

    # simplify for example
    print(f'got command {command}')
    book = Book(**dataclasses.asdict(command))

    # returning the event leads to emitting it by the dispatcher,
    # another way to do that will be described in classes example
    return BookCreated(book=book)


@dp.events.register(BookCreated, storage=Singleton["storage": get_storage])
async def book_created_handler(event: BookCreated, storage: dict) -> None:
    """Handle book creation"""
    print(f'got event {event}')
    book = event.book
    storage[book.title] = book


@dp.queries.register(BookQuery, storage=Singleton["storage": get_storage])
# ensure you did not reassign the type of provider (f.e. from Singleton to Factory)
async def book_query_handler(query: BookQuery, storage: dict) -> list[Book]:
    """Handle book query"""
    # find books in storage...
    books = [book for title, book in storage.items() if title == query.title]
    print(f"got query {query}, found {len(books)} books")
    return books


@dp.events.register('on startup')
# # you can also use strings for message identifiers
async def on_start() -> None:
    """Handlers without message argument also can be used"""
    # do something on startup
    print('application started')


async def main() -> None:
    """Application entrypoint"""
    logging.basicConfig(level=logging.DEBUG)

    dp.start()  # start router's engines

    await dp.handle('on startup')
    await dp.handle(
        CreateBook(title="Philosopher's Stone", author="J. K. Rowling", year=1997),
        key='createBook',
    )
    books = await dp.handle(
        BookQuery(title="Philosopher's Stone"), wait_for_response=True,
    )
    print(f"found {books=}")

    await signals.wait_for_shutdown()


if __name__ == '__main__':
    asyncio.run(main())



import abc
import asyncio
import typing as t
from logging import getLogger

from iambus.core.api.isolation import IsolationManager
from iambus.core.api.typedef import RequestObject
from iambus.core.api.typing import MapReturnType, MessageMapType, MessageType, WrappedHandler

logger = getLogger(__name__)


class BaseQueueWorker(metaclass=abc.ABCMeta):
    """Base queue worker class."""

    def __init__(self, engine: "AbstractEngine", key: t.Optional[t.AnyStr] = None):
        self._engine = engine
        self._name = f'{self._engine.name} (key {key})'
        self._queue = asyncio.Queue[RequestObject]()
        self._started = False

    def tell(self) -> int:
        """Return size of the queue."""
        return self._queue.qsize()

    @abc.abstractmethod
    def start(self) -> None:
        """Start the worker."""

    @abc.abstractmethod
    async def put(self, request: RequestObject) -> None:
        """Put the message to the queue."""

    async def _worker(self):
        while True:
            try:
                logger.debug(f'{self._name} waiting for messages')
                request = await self._queue.get()
                logger.debug(f'{self._name} got message {request.message}')
                handler = self._engine.map.find(request.message)

                try:
                    response = await self._engine.handle(
                        handler,
                        message=request.message,
                        key=request.key,
                    )
                    if request.wait_for_response:
                        request.callback.set_result(response)
                except Exception as exc:
                    if request.wait_for_response:
                        request.callback.set_exception(exc)
                    else:
                        await self._engine.error_handler(exc)

            except asyncio.CancelledError:
                break

            except Exception as e:
                await self._engine.error_handler(e)


class QueueWorker(BaseQueueWorker):
    """Single key-bound queue worker."""

    def start(self):
        """Start the worker."""
        asyncio.create_task(self._worker())
        self._started = True

    async def put(self, request: RequestObject) -> None:
        """Put the message to the queue."""
        if not self._started:
            self.start()

        await self._queue.put(request)


class MultiQueueWorker:
    """Multiple key-bound queue worker."""

    def __init__(self, engine: "AbstractEngine"):
        self._engine = engine
        self._queues = []

    def get_queue(self):
        """Return min tasks queue."""
        return min(self._queues, key=lambda q: q.tell())

    def start(self, workers: int):
        """Start the worker."""
        self._queues = [QueueWorker(self._engine) for _ in range(workers)]

    async def put(self, request: RequestObject) -> None:
        """Put message to the queue."""
        await self.get_queue().put(request)


class AbstractEngine(t.Generic[MessageMapType], metaclass=abc.ABCMeta):
    """Engine protocol.

    Implement:
      - handle
    """

    def __init__(self, message_map: MessageMapType):
        self.map = message_map
        self._default = MultiQueueWorker(self)
        self._topics: dict[t.AnyStr, QueueWorker] = {}
        self._isolation = IsolationManager()

        self._workers: int = 0
        self._name: str = self.__class__.__name__
        self._stop_event = asyncio.Event()
        self._started = False

    name = property(lambda self: self._name)

    @property
    def is_started(self) -> bool:
        """Return True if the engine is started."""
        return self._started

    async def put_to_queue(self, request: RequestObject) -> t.Awaitable[...]:
        """Put the message to the queue."""
        if request.wait_for_response is not True:
            request.callback.set_result(None)

        async with self._isolation.get_lock(ikey=request.isolation_key()):
            if key := request.key:
                # for each key create a queue worker
                await self._topics.setdefault(key, QueueWorker(self, key)).put(request)
            else:
                # put to default if no key presente
                await self._default.put(request)

        return await request.callback

    async def error_handler(self, error: str | Exception) -> None:  # noqa
        """Error handler"""
        logger.exception(error)

    @abc.abstractmethod
    async def handle(
        self,
        handler: MapReturnType,
        /,
        message: MessageType,
        key: t.Optional[t.AnyStr] = None,
    ) -> t.Any:
        """Handler the message. Optionally can return new messages."""

    async def handle_one(
        self,
        handler: WrappedHandler,
        message: MessageType,
        key: t.Optional[t.AnyStr] = None,
    ):
        """Handle one message by one handler."""
        response = await handler.handle(message)
        asyncio.ensure_future(self.handle_side_events(handler, key=key))
        return response

    async def handle_side_events(
        self,
        handler: WrappedHandler,
        key: t.Optional[t.AnyStr] = None,
    ) -> None:
        """Handle side handler event"""
        for event in await handler.dump_events():
            await self.put_to_queue(
                RequestObject(message=event, key=key, wait_for_response=False),
            )

    def start(self, workers: int = 3) -> None:
        """Start the engine."""
        self._default.start(workers)
        self._started = True



import asyncio
from concurrent.futures import Future


def handler_repr(handler):  # pragma: no cover
    """Return a string representation of the given handler."""
    cls = (
        handler.__self__.__class__.__name__ if hasattr(handler, "__self__")
        else handler.__class__.__name__ if hasattr(handler, "__class__")
        else ""
    )

    div = "." if cls else ""
    handler_name = f'{handler.__module__}.{cls}{div}'

    if hasattr(handler, '__name__'):
        handler_name += handler.__name__

    else:
        if "handle" in dir(handler):
            handler_name += 'handle'
        else:
            print(f'unknown handler {handler=!r}')
            print(dir(handler))

    return handler_name


def get_async_result(coro):
    """Waits async results in sync environment, decorators f.e."""
    loop = asyncio.get_event_loop()

    if not loop.is_running():
        return loop.run_until_complete(coro)

    future = Future()

    def _callback():
        asyncio.ensure_future(coro).add_done_callback(
            lambda task: future.set_result(task.result())
        )

    loop.call_soon_threadsafe(_callback)
    return future.result()


def loop_time() -> float:
    """Return the current loop time."""
    return asyncio.get_event_loop().time()



import asyncio
import typing as t
from dataclasses import dataclass, field

from iambus.core.api.isolation import IsolationKey
from iambus.core.api.typing import MessageType, make_key

UNHANDLED = object()


def _future_cb() -> asyncio.Future:
    loop = asyncio.get_event_loop()
    return loop.create_future()


@dataclass(kw_only=True, slots=True)
class RequestObject:
    """Represents a request object."""
    message: MessageType
    key: t.Optional[t.AnyStr] = None
    wait_for_response: bool = False
    callback: asyncio.Future = field(default_factory=_future_cb)

    def isolation_key(self):
        """Isolation key of the request object."""
        return IsolationKey(obj_key=self.key, event_key=make_key(self.message))


@dataclass
class ResponseObject:
    """Represents a response object."""
    pass



import abc
import inspect
import typing as t
from collections import defaultdict
from logging import getLogger

from iambus.core import exceptions as exc
from iambus.core import helpers

logger = getLogger(__name__)

ProvidedType = t.Any
ProviderFn = t.Callable[[], t.Awaitable[ProvidedType] | ProvidedType]
ProviderItem = dict[str, ProviderFn | t.Optional[ProvidedType]]


class Provider(metaclass=abc.ABCMeta):
    """Base dependency provider."""

    providers: dict[str, ProviderItem] = defaultdict(dict)

    @classmethod
    def _func_call(cls, fn: ProviderFn) -> ProvidedType:
        """Call func"""

        if inspect.iscoroutinefunction(fn):
            return helpers.get_async_result(fn())

        return fn()

    @classmethod
    def _get_provider(cls, name: str) -> ProviderItem:
        if (provided := cls.providers.get(name, None)) is None:
            raise KeyError(name)

        return provided

    @classmethod
    @abc.abstractmethod
    def get_provided(cls, provider: ProviderItem) -> ProvidedType:
        """Return provided object"""

    @classmethod
    def call(cls, name: str) -> ProvidedType:
        """Call the provider."""
        provided = cls._get_provider(name)
        return cls.get_provided(provided)

    def __class_getitem__(cls, item: ProviderFn):
        if isinstance(item, slice):
            name, fn, *_ = str(item.start), item.stop
        else:
            raise TypeError(
                f"{type(item)} not supported for {cls.__name__} generic, "
                f"use slice syntax: fn_name: fn"
            )

        if name in cls.providers:
            if (cls_type := cls.providers[name]['type']) != cls.__name__:
                raise exc.ImproperlyConfigured(
                    reason=f'attempt to reassign provider class from {cls_type} to {cls.__name__}'
                )
            return cls

        cls.providers[name] = {
            "item": fn,
            "raw": True,
            "type": cls.__name__,
        }
        return cls


class Singleton(Provider):
    """Singleton dependency provider."""

    @classmethod
    def get_provided(cls, provider: ProviderItem) -> ProvidedType:
        provided = provider['item']

        if provider['raw'] is False:
            return provided

        provided = provider['item'] = cls._func_call(provided)
        provider['raw'] = False

        return provided


class Factory(Provider):
    """Factory dependency provider."""

    @classmethod
    def get_provided(cls, provider: ProviderItem) -> ProvidedType:
        """Call the provider."""
        return cls._func_call(provider['item'])



import asyncio
import dataclasses
import logging

from examples.books.messages import CreateBook, BookCreated, BookQuery
from examples.books.models import Book
from iambus import dispatcher as dp
from iambus.core import signals


# somewhere in your code
async def create_book_handler(command: CreateBook) -> BookCreated:
    """Create book"""

    # simplify for example
    print(f'got command {command}')
    book = Book(**dataclasses.asdict(command))

    # returning the event leads to emitting it by the dispatcher,
    # another way to do that will be described in classes example
    return BookCreated(book=book)


async def book_created_handler(event: BookCreated, storage: dict) -> None:
    """Handle book creation"""
    print(f'got event {event}')
    book = event.book
    storage[book.title] = book


async def book_query_handler(query: BookQuery, storage: dict) -> list[Book]:
    """Handle book query"""
    # find books in storage...
    books = [book for title, book in storage.items() if title == query.title]
    print(f"got query {query}, found {len(books)} books")
    return books


async def on_start() -> None:
    """Handlers without message argument also can be used"""
    # do something on startup
    print('application started')


async def main() -> None:
    """Application entrypoint"""
    logging.basicConfig(level=logging.DEBUG)

    # simple dictionary storage for example
    storage = {}

    dp.commands.bind(CreateBook, handler=create_book_handler, response_event=BookCreated)
    dp.events.bind(BookCreated, handler=book_created_handler, storage=storage)
    dp.queries.bind(BookQuery, handler=book_query_handler, storage=storage)

    # you can also use strings for message identifiers
    dp.events.bind('on startup', handler=on_start)

    dp.start()  # start router's engines

    await dp.events.handle('on startup')
    await dp.handle(
        CreateBook(title="Philosopher's Stone", author="J. K. Rowling", year=1997),
        key='createBook', wait_for_response=True,
    )
    await asyncio.sleep(0.2)  # ! CreateBook command will be executed without blocking
    books = await dp.handle(
        BookQuery(title="Philosopher's Stone"), wait_for_response=True,
        key='createBook',
    )
    print(f'found {books=}')

    await signals.wait_for_shutdown()


if __name__ == '__main__':
    asyncio.run(main())



import typing as t

from iambus.core.api.typing import (
    EventRouterType,
    RequestRouterType, MessageType,
)


@t.runtime_checkable
class DispatcherProtocol(t.Protocol[EventRouterType, RequestRouterType]):
    """DispatcherProtocol protocol."""

    @property
    def events(self) -> EventRouterType:
        """Return events proxy"""

    @property
    def commands(self) -> RequestRouterType:
        """Return commands proxy"""

    @property
    def queries(self) -> RequestRouterType:
        """Return queries proxy"""

    async def handle(
        self,
        message: MessageType,
        key: t.Optional[t.AnyStr],
        wait_for_response: bool,
    ) -> t.Awaitable[...]:
        """Handle a message"""



import typing as t

from iambus.base.engine.events import EventEngine
from iambus.base.maps import RequestHandlerMap
from iambus.core.api.engine import AbstractEngine
from iambus.core.api.typedef import RequestObject
from iambus.core.api.typing import MessageType, WrappedHandler


class RequestEngine(AbstractEngine[RequestHandlerMap]):

    def __init__(self, event_engine: EventEngine, message_map: RequestHandlerMap):
        super().__init__(message_map=message_map)
        self._event_engine = event_engine

    async def handle(
        self,
        handler: WrappedHandler,
        /,
        message: MessageType,
        key: t.Optional[t.AnyStr] = None,
    ):
        return await self.handle_one(handler, message, key)

    async def handle_side_events(
        self,
        handler: WrappedHandler,
        key: t.Optional[t.AnyStr] = None,
    ) -> None:
        """Handle side handler events"""
        for event in await handler.dump_events():
            await self._event_engine.put_to_queue(
                RequestObject(message=event, key=key, wait_for_response=False),
            )



import dataclasses
from typing import ClassVar, Any


# defining domain model
@dataclasses.dataclass
class Book:
    """Book model"""
    title: str
    author: str
    year: int


# defining application message models
@dataclasses.dataclass(frozen=True, slots=True)
class Message:
    """Application Message model"""
    __dataclass_fields__: ClassVar[dict[str, Any]]


@dataclasses.dataclass(frozen=True, slots=True)
class Event(Message):
    """Application Event model"""


@dataclasses.dataclass(frozen=True, slots=True)
class Command(Message):
    """Application Command model"""


@dataclasses.dataclass(frozen=True, slots=True)
class Query(Message):
    """Application Query model"""



import dataclasses

from examples.books.models import Book, Command, Query, Event


# defining simple messages: command, event and query
@dataclasses.dataclass(frozen=True, slots=True)
class BookCreated(Event):
    """Book created event"""
    book: Book


@dataclasses.dataclass(frozen=True, slots=True)
class CreateBook(Book, Command):
    """Create book Command"""


@dataclasses.dataclass(frozen=True, slots=True)
class BookQuery(Query):
    """Book Query model"""
    title: str



import asyncio
import signal
from logging import getLogger

prepare_shutdown = asyncio.Event()
logger = getLogger(__name__)
logger.setLevel("DEBUG")


def _handle_shutdown(sig):
    logger.debug(f"received {sig.name}, shutting down")
    prepare_shutdown.set()


def _create_listeners():
    loop = asyncio.get_running_loop()

    for sig in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(
            sig, _handle_shutdown, sig
        )


def setup():
    """Setup library graceful shutdown."""
    _create_listeners()


async def wait_for_shutdown():
    """Alias to the shutdown event"""
    await prepare_shutdown.wait()



import abc
from abc import ABC
from typing import Self

from iambus.core.api.handlers import AbstractHandler
from iambus.core.api.typing import MessageType


class HandlerMeta(abc.ABCMeta):
    """Abstract Handler Meta"""

    def __call__(cls, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        obj._events = []
        return obj

    def __new__(mcs, name: str, bases: tuple, namespace: dict, /, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        # cls._events = []
        cls.dump_events = mcs._dump_events
        cls.add_event = mcs._add_event

        return cls

    async def _dump_events(self: Self) -> list[MessageType]:
        events = self._events
        self._events = []
        return events

    async def _add_event(self: Self, event: MessageType) -> None:
        self._events.append(event)


class PyBusAbstractHandler(AbstractHandler, ABC, metaclass=HandlerMeta):
    """
    PyBusHandler abstract base class.

    To implement:
      - handle

    """



import asyncio
import typing as t
from logging import getLogger

from iambus.base.maps import EventHandlerMap
from iambus.core.api.engine import AbstractEngine
from iambus.core.api.typedef import UNHANDLED
from iambus.core.api.typing import MessageType, WrappedHandler

logger = getLogger(__name__)


class EventEngine(AbstractEngine[EventHandlerMap]):

    async def handle(
        self,
        handlers: frozenset[WrappedHandler],
        /,
        message: MessageType,
        key: t.Optional[t.AnyStr] = None,
    ):
        if not handlers:
            return UNHANDLED

        await asyncio.gather(
            *(self.handle_one(handler, message, key) for handler in handlers),
        )



import inspect
import typing as t
from logging import getLogger
from typing import Optional

from iambus.base.routers.eventrouter import EventRouter
from iambus.base.routers.requestrouter import RequestRouter
from iambus.core import signals
from iambus.core.api.dispatcher import DispatcherProtocol
from iambus.core.api.typedef import UNHANDLED
from iambus.core.api.typing import MessageType

logger = getLogger('iambus.dispatcher')


class Dispatcher(DispatcherProtocol[EventRouter, RequestRouter]):
    """Base dispatcher protocol implementation."""
    _commands = None
    _queries = None

    def __init__(
        self,
        events_router_cls: type[EventRouter] = EventRouter,
        commands_router_cls: Optional[type[RequestRouter]] = None,
        queries_router_cls: Optional[type[RequestRouter]] = None,
        listen_for_signals: bool = True,
        engine_workers: int = 1,
    ) -> None:

        self._engine_workers = engine_workers
        self._routers = []

        if events_router_cls:
            self._events = events_router_cls(self)
            self._routers.append(self._events)

        if commands_router_cls:
            self._commands = commands_router_cls(self)
            self._routers.append(self._commands)

        if queries_router_cls:
            self._queries = queries_router_cls(self)
            self._routers.append(self._queries)

        self._listen_for_signals = listen_for_signals
        self._started = False

    @property
    def events(self):  # pragma: no cover
        if self._events is None:
            logger.info('attaching default event router')
            self._events = EventRouter(self)
            self._routers.append(self._events)

        return self._events

    @property
    def commands(self):  # pragma: no cover
        if self._commands is None:
            logger.info('attaching default command router')
            self._commands = RequestRouter(self)
            self._routers.append(self._commands)

        return self._commands

    @property
    def queries(self):  # pragma: no cover
        if self._queries is None:
            logger.info('attaching default query router')
            self._queries = RequestRouter(self)
            self._routers.append(self._queries)

        return self._queries

    def start(self) -> None:
        """Start the dispatcher."""

        if not any([
            self.events is not None,
            self.queries is not None,
            self.commands is not None
        ]):
            return logger.warning('no handlers registered')

        if self._events is not None:
            self.events.setup(workers=self._engine_workers)

        if self._commands is not None:
            self.commands.setup(workers=self._engine_workers)

        if self._queries is not None:
            self.queries.setup(workers=self._engine_workers)

        if self._listen_for_signals:
            signals.setup()

        logger.debug(f'{self.__class__.__name__} started.')
        self._started = True

    async def handle(
        self,
        message: MessageType,
        key: t.Optional[t.AnyStr] = None,
        wait_for_response: bool = False,
    ):
        """Try to handle message by event, command and query routers respectively.

        :param message: The message to handle.
        :param key: The key to use for the message partition.
        :param wait_for_response: If True, wait for a response.
        """
        for router in self._routers:
            response = await router.handle(
                message=message,
                key=key,
                wait_for_response=wait_for_response,
            )
            if response is not UNHANDLED:
                if inspect.isawaitable(response):
                    return await response
                return response

        logger.error(f'{message} was not handled.')


default_dispatcher = Dispatcher()



import asyncio
import contextlib
from collections import defaultdict
from dataclasses import dataclass
from typing import AnyStr, Optional

from iambus.core.api.typing import TypeKey


@dataclass(frozen=True, slots=True, kw_only=True)
class IsolationKey:
    """Event identify key"""
    event_key: TypeKey
    """Event key: type of the message or value if message is string"""
    obj_key: Optional[AnyStr] = None
    """Optional key for the partition message"""


class IsolationManager:
    """Lock manager for events"""
    __slots__ = ("_locks",)

    def __init__(self):
        self._locks: dict[IsolationKey, asyncio.Lock] = defaultdict(asyncio.Lock)

    @contextlib.asynccontextmanager
    async def get_lock(self, ikey: IsolationKey):
        """Acquire lock in context"""
        async with self._locks[ikey]:
            yield



from iambus.base.dispatcher import Dispatcher as Dispatcher
from iambus.base.dispatcher import default_dispatcher as dispatcher
from iambus.base.engine.events import EventEngine as EventEngine
from iambus.base.engine.requests import RequestEngine as RequestEngine
from iambus.base.handlers.abstract import PyBusAbstractHandler as AbstractHandler
from iambus.base.maps import EventHandlerMap as EventHandlerMap
from iambus.base.maps import RequestHandlerMap as RequestHandlerMap
from iambus.base.routers.eventrouter import EventRouter as EventRouter
from iambus.base.routers.requestrouter import RequestRouter as RequestRouter
from iambus.core.dependency.providers import Singleton as Singleton, Factory as Factory

__all__ = [
    "AbstractHandler",
    "Dispatcher",
    "EventHandlerMap",
    "EventEngine",
    "EventRouter",
    "Factory",
    "RequestHandlerMap",
    "RequestEngine",
    "RequestRouter",
    "Singleton",
    "dispatcher",
]
